#ifndef _MYSQLFORMATER_H_INCLUDED_
#define _MYSQLFORMATER_H_INCLUDED_
///////////////////////////////////////////////////////////////////////////////
//                                                         
// MySQLFormater.h  
// ---------------
// Exort to mysql format
//
// Design and Implementation by Bjoern Lemke
//
// (C)opyright 2015 Bjoern Lemke
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; see the file COPYING.  If not, write to
// the Free Software Foundation, 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.
//
// INTERFACE MODULE
//
// Class: MySQLFormater
// 
// Description:
//
//
///////////////////////////////////////////////////////////////////////////////


#include <cego/CegoNet.h>
#include <cego/CegoDatabaseFormater.h>

class MySQLFormater : public CegoDatabaseFormater {
    
public:
    
    MySQLFormater(CegoNet *pCegoNet, const Chain& sqlfile);
    ~MySQLFormater();

    Chain formatProcedure();
    
    void exportTable();
    void exportProcedure();
    void exportView();
    void exportIndex();
    void exportBtree();
    void exportConstraint();
    
    // format method implementation
    Chain formatProcedure(const Chain& procName, CegoProcedure::ProcType procType, CegoDataType returnType, int returnTypeLen, int returnTypeDim, CegoProcBlock* pBlock);
    Chain formatProcVar(const Chain& varName, CegoProcVar::VarType varType, CegoDataType dt, int dtlen, int dtdim); 
    Chain formatProcBlock(const ListT<CegoProcVar>& varList, const ListT<CegoProcStmt*>& stmtList,  const ListT<CegoProcException*>& exceptionList);

    Chain formatProcThrowStmt(CegoExpr* pExpr);
    Chain formatProcAssignStmt(CegoProcAssignStmt::AssignmentMode mode, const Chain paramName, CegoExpr* pExpr);
    Chain formatProcBlockStmt(CegoProcBlock* pBlock);
    Chain formatProcCursorCreateStmt(const Chain& name, CegoSelect* pSelect);
    Chain formatProcCursorCloseStmt(const Chain& name);
    Chain formatProcIfStmt(const ListT<CegoProcCond*>& condList, const ListT<CegoProcBlock*>& ifBlockList);
    Chain formatProcNoopStmt();
    Chain formatProcQueryStmt(CegoQuery *pQuery);
    Chain formatProcReturnStmt(CegoExpr *pExpr);
    Chain formatProcWhileStmt(CegoProcCond* pCond, CegoProcBlock* pBlock);
    Chain formatProcCond(CegoProcCond::ConditionMode mode, CegoProcCond* pCond, CegoProcPred* pPred);
    Chain formatProcPred(CegoProcPred::PredicateMode mode, CegoExpr* pExpr1, CegoExpr* pExpr2, CegoComparison comp, CegoProcCond* pCond, bool isNegate);
    Chain formatProcException(CegoException excep);
    
    Chain formatExpr(CegoExpr::ExpType expType, CegoExpr* pExpr, CegoTerm* pTerm);
    Chain formatTerm(CegoTerm::TermType termType, CegoTerm* pTerm, CegoFactor* pFactor);
    Chain formatFactor(CegoFactor::FacType facType, const CegoFieldValue& fv, CegoAttrDesc* pAttrDesc, const Chain varName,
		       CegoExpr* pExpr, CegoProcFetch *pFetch, CegoAggregation* pAggr, CegoFunction* pFunction,
		       CegoSelect* pSelect, CegoCaseCond* pCaseCond );
    
    Chain formatSelect(CegoSelect::SelectMode mode, bool isDistinct, const ListT<CegoExpr*>& exprList,
		       const ListT<CegoContentObject*>& coList, CegoPredDesc *pPred, ListT<CegoAttrDesc*> *pGroupList, CegoHavingCond* pHaving, ListT<CegoExpr*> *pOrderList, int rowLimit, CegoSelect* pUnionSelect);
    Chain formatProcFetch(const Chain& curName, const ListT<Chain>& varList);
    
    Chain formatHavingCond(CegoHavingCond::ConditionMode mode, CegoHavingCond* pCond, CegoHavingDesc* pDesc);
    Chain formatHavingDesc(CegoExpr *pAggExpr, CegoExpr* pExpr, CegoComparison comp);

    
    Chain formatPred(CegoPredDesc::CompMode mode, CegoExpr* pExpr1, CegoExpr *pExpr2, CegoExpr* pExpr3, ListT<CegoExpr*>& exprList, CegoComparison comp, const Chain& pattern, CegoCondDesc* pCond, CegoPredDesc* pPred, CegoSelect *pSelect);

    Chain formatQuery(CegoQuery::QueryMode mode, const Chain& tableName, const Chain& tableAlias, const ListT<CegoField>& schema, const ListT<ListT<CegoExpr*> >& exprListArray,
		      const ListT<CegoExpr*>& updList, const ListT<CegoReturnVar*>& retVarList, const ListT<CegoAlterDesc>& alterList, CegoPredDesc *pPred, CegoSelect *pSelect,
		      const Chain& objName, CegoObject::ObjectType objType, const Chain& newObjName);
        
    Chain formatAttr(const Chain& tableName, const Chain& attrName);
    Chain formatFunction(CegoFunction::FunctionType type, const ListT<CegoExpr*>& exprList, const Chain& funcName, const Chain& counterId);
    Chain formatAggregation(CegoAggregation::AggType type, CegoExpr *pExpr, bool isDistinct);
    Chain formatCaseCond(const ListT<CegoPredDesc*>& predList, const ListT<CegoExpr*>& exprList, CegoExpr* elseExpr);

    Chain formatReturnValue(const Chain& varName, CegoExpr* pExpr);
    Chain formatAlter(CegoAlterDesc::AlterType type, const CegoField& f, const Chain& attrName);
    Chain formatFieldValue(CegoDataType type, void *pV, int len );

    Chain formatTrigger(const Chain& triggerName,
			bool isBefore,
			bool isOnInsert,
			bool isOnUpdate,
			bool isOnDelete,
			const Chain& tableName,				
			CegoProcBlock* pBlock);
	
private:

    Chain formatIndent() const;
    
    CegoNet* _pCegoNet;
    File* _pDF;
    int _indentCount;
    bool _isPlain;
    Chain _fetchStmt;

    ListT<Chain> _declareList;
    ListT<Chain> _setList;
    
};

#endif














