///////////////////////////////////////////////////////////////////////////////
//                                                         
// OracleFormater.cc  
// ----------------                                                        
// Formatting Oracle SQL statements
//
// Design and Implementation by Bjoern Lemke
//     
// (C)opyright 2015 Bjoern Lemke
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; see the file COPYING.  If not, write to
// the Free Software Foundation, 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.
//
// IMPLEMENTATION MODULE
//
// Class: OracleFormater
// 
// Description: 
//
//
///////////////////////////////////////////////////////////////////////////////

#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include "OracleFormater.h"

#include "cego/CegoXMLdef.h"
#include "cego/CegoAction.h"
#include "cego/CegoProcedure.h"
#include "lfcbase/BigInteger.h"
#include "lfcbase/BigDecimal.h"
#include "lfcbase/Datetime.h"


#define ORACLEDATEFORMAT "%Y-%m-%d %H:%M:%S"

OracleFormater::OracleFormater(CegoNet *pCegoNet, const Chain& sqlfile) : CegoDatabaseFormater()
{
    if ( pCegoNet )
    {
	_pCegoNet = pCegoNet;
	_pDF = new File(sqlfile);
	_pDF->open(File::WRITE);
    }
    else
    {
	throw Exception(EXLOC, "Cego database connection required for export"); 
    }
}

OracleFormater::~OracleFormater()  
{
    _pDF->close();
    delete _pDF;	
}

void OracleFormater::exportTable()
{
    long res;
    res = _pCegoNet->doQuery(Chain("list table;"));
    
    ListT<CegoField> schema;
    _pCegoNet->getSchema(schema);
    
    ListT<Chain> tableList;
    ListT<CegoFieldValue> fvl;
    while ( _pCegoNet->fetchData(schema, fvl)  )
    {

	CegoFieldValue *pFV = fvl.First();
	Chain tableName = pFV->valAsChain();

	cout << "Got table " << tableName << endl;
	
	tableList.Insert(tableName);
	fvl.Empty();	
    }
    
    Chain *pTable = tableList.First();
    while ( pTable ) 
    {
	res = _pCegoNet->doQuery(Chain("desc table ") + *pTable + Chain(";"));

	Chain dropStmt;

	dropStmt=Chain("DROP TABLE IF EXISTS ") + *pTable + Chain(";\n");
	_pDF->writeChain(dropStmt);
	
	Chain createStmt=Chain("CREATE TABLE ") + *pTable + Chain("(");

	ListT<CegoField> schema;
	_pCegoNet->getSchema(schema);
    
	ListT<CegoFieldValue> fvl;
	bool isFirst=true;
	while ( _pCegoNet->fetchData(schema, fvl)  )
	{
	    if ( isFirst == false )
		createStmt += Chain(", ");
	    isFirst=false;

	    CegoFieldValue *pFV = fvl.First();
	    Chain attrName = pFV->valAsChain();
	    pFV = fvl.Next();
	    Chain attrType = pFV->valAsChain();
	    pFV = fvl.Next();
	    Chain attrSize = pFV->valAsChain();
	    pFV = fvl.Next();
	    Chain attrDef = pFV->valAsChain();
	    pFV = fvl.Next();
	    Chain attrNull = pFV->valAsChain();

	    createStmt += attrName + Chain(" ");
	    
	    if ( attrType == Chain(XML_STRING_VALUE) )
		createStmt += Chain("VARCHAR") + Chain("(") + attrSize + Chain(")");
	    else if ( attrType == Chain(XML_INT_VALUE) )
		createStmt += Chain("INT");
	    // else if ( attrType == ....
	    
	    if ( attrDef != Chain(XML_NULL_VALUE))
		createStmt += Chain(" DEFAULT ") + attrDef;
	    if ( attrNull == Chain("no"))
		createStmt += Chain(" NOT NULL");
	    fvl.Empty();	
	}

	createStmt += Chain(");\n");
	_pDF->writeChain(createStmt);


	// dump table content
	res = _pCegoNet->doQuery(Chain("select * from ") + *pTable + Chain(";"));
	    
	schema.Empty();
	fvl.Empty();
	
	_pCegoNet->getSchema(schema);
	
	while ( _pCegoNet->fetchData(schema, fvl)  )
	{		
	    Chain insertStmt=Chain("INSERT INTO ") + *pTable + Chain(" VALUES (");
	    
	    CegoFieldValue *pFV = fvl.First();
	    while ( pFV )
	    {
		Chain val = pFV->toChain();
		insertStmt += val;
		pFV = fvl.Next();
		
		if ( pFV )
		    insertStmt += Chain(", ");
	    }
	    insertStmt += Chain(");\n");
	    _pDF->writeChain(insertStmt);
	    fvl.Empty();
	}

	pTable = tableList.Next();
    }
}

void OracleFormater::exportProcedure()
{
    long res;
    res = _pCegoNet->doQuery(Chain("list procedure;"));
    
    ListT<CegoField> schema;
    _pCegoNet->getSchema(schema);
    
    ListT<Chain> procList;
    ListT<CegoFieldValue> fvl;
    while ( _pCegoNet->fetchData(schema, fvl)  )
    {
	CegoFieldValue *pFV = fvl.First();
	Chain procName = pFV->valAsChain();
	procList.Insert(procName);
	fvl.Empty();	
    }

    _pDF->writeChain(Chain("DELIMITER //\n"));
    
    Chain *pProcName = procList.First();
    while ( pProcName ) 
    {
	res = _pCegoNet->doQuery(Chain("show procedure ") + *pProcName + Chain(";"));

	ListT<CegoField> schema;
	_pCegoNet->getSchema(schema);
    
	ListT<CegoFieldValue> fvl;
	Chain procText;
	while ( _pCegoNet->fetchData(schema, fvl)  )
	{
	    CegoFieldValue *pFV = fvl.First();
	    procText = pFV->valAsChain();    
	    fvl.Empty();	
	}
	
	CegoAction* pA = new CegoAction(0,0);
	pA->cleanUp();

	/*
	Chain procLoad = Chain("load procedure copytab(x in int, y in string(30)) return int \
begin \
   var a int; \
   :a = ( select b from cXX ); \
   insert into t2 select a, b from t1; \
   return :a; \
end;");
	*/

	Chain procLoad = Chain("load ") + procText;
	pA->setCommandChain(procLoad);
	pA->parse();

	CegoProcedure* pProc = pA->getProcedure(false);

	Chain dropStmt;
	if ( pProc->getProcType() == CegoProcedure::FUNCTION )
	    dropStmt=Chain("DROP FUNCTION IF EXISTS ") + *pProcName + Chain("\n");
	else
	    dropStmt=Chain("DROP PROCEDURE IF EXISTS ") + *pProcName + Chain("\n");
	
	_pDF->writeChain(dropStmt);
	_pDF->writeChain(Chain("//\n"));
	_pDF->writeChain(pProc->dbFormat(this));
	_pDF->writeChain(Chain("//\n"));
	
	delete pA;
	
	pProcName = procList.Next();
    }

}


void OracleFormater::exportView()
{
    long res;
    res = _pCegoNet->doQuery(Chain("list view;"));
    
    ListT<CegoField> schema;
    _pCegoNet->getSchema(schema);
    
    ListT<Chain> viewList;
    ListT<CegoFieldValue> fvl;
    while ( _pCegoNet->fetchData(schema, fvl)  )
    {
	CegoFieldValue *pFV = fvl.First();
	Chain viewName = pFV->valAsChain();
	viewList.Insert(viewName);
	fvl.Empty();	
    }
    
    Chain *pViewName = viewList.First();
    while ( pViewName ) 
    {
	res = _pCegoNet->doQuery(Chain("show view ") + *pViewName + Chain(";"));

	ListT<CegoField> schema;
	_pCegoNet->getSchema(schema);
    
	ListT<CegoFieldValue> fvl;
	Chain viewText;
	while ( _pCegoNet->fetchData(schema, fvl)  )
	{
	    CegoFieldValue *pFV = fvl.First();
	    viewText = pFV->valAsChain();    
	    fvl.Empty();	
	}
	
	CegoAction* pA = new CegoAction(0,0);
	pA->cleanUp();

	Chain viewLoad = Chain("load ") + viewText;
	pA->setCommandChain(viewLoad);
	pA->parse();


	CegoSelect* pSelect = pA->getSelect(false);
		
	Chain dropStmt;
       
	dropStmt=Chain("DROP VIEW IF EXISTS ") + *pViewName + Chain(";\n");
	_pDF->writeChain(dropStmt);

	_pDF->writeChain(pSelect->dbFormat(this));
	
	delete pA;
	
	pViewName = viewList.Next();
    }

}


Chain OracleFormater::formatProcedure(const Chain& procName, CegoProcedure::ProcType procType, CegoDataType returnType, int returnTypeLen, int returnTypeDim, CegoProcBlock* pBlock)
{

    Chain s;

    if ( procType == CegoProcedure::FUNCTION )
	s = Chain("CREATE FUNCTION ");
    else
	s = Chain("CREATE PROCEDURE ");
    
    s += procName + Chain("(");
    CegoProcVar *pVar = pBlock->getVarList().First();
    while ( pVar )
    {
	if ( pVar->getVarType() == CegoProcVar::BLOCKVAR )
	    pVar = 0;
	else
	{
	    s += pVar->dbFormat(this);
	    pVar = pBlock->getVarList().Next();
	    	    
	    if ( pVar )
	    {
		if ( pVar->getVarType() != CegoProcVar::BLOCKVAR )
		s += Chain(",\n     ");
	    }
	}
    } 

    s += Chain(")");
    if ( procType == CegoProcedure::FUNCTION )
    {
	s += Chain(" RETURNS");
	switch ( returnType )
	{
	case INT_TYPE:
	    s += Chain(" INT");
	    break;
	case LONG_TYPE:
	    s += Chain(" LONG");
	    break;
	case VARCHAR_TYPE:
	    s += Chain(" VARCHAR(");
	    s += Chain(returnTypeLen);
	    s += Chain(")");
	    break;
	case BOOL_TYPE:
	    s += Chain(" BOOL");
	    break;
	case DATETIME_TYPE:
	    s += Chain(" DATETIME");
	    break;
	case BIGINT_TYPE:
	    s += Chain(" bigint(");
	    s += Chain(returnTypeLen);
	    s += Chain(")");
	    break;
	case FLOAT_TYPE:
	    s += Chain(" float");
	    break;
	case DOUBLE_TYPE:
	    s += Chain(" double");
	    break;
	case DECIMAL_TYPE:
	    s += Chain(" decimal(");
	    s += Chain(returnTypeLen);
	    s += Chain(")");
	    break;
	case SMALLINT_TYPE:
	    s += Chain(" smallint");
	    break;
	case TINYINT_TYPE:
	    s += Chain(" tinyint");
	    break;	  
	case NULL_TYPE:
	    s += Chain(" null");
	    break;
	case BLOB_TYPE:
	case CLOB_TYPE:
	case PAGEID_TYPE:
	    throw Exception(EXLOC, Chain("Type not supported as return type"));
	}
    }

    s += Chain(" DETERMINISTIC \n");
    s += Chain("BEGIN\n");

    Chain indent = Chain("   ");

    s += pBlock->dbFormat(this);
    
    s += Chain("END\n");
    return s;

}

Chain OracleFormater::formatProcVar(const Chain& varName, CegoProcVar::VarType varType, CegoDataType dt, int dtlen, int dtdim)
{
    Chain s = varName;
    
    switch ( dt )
    {
    case INT_TYPE:
	s += Chain(" INT");
	break;
    case LONG_TYPE:
	s += Chain(" LONG");
	break;
    case VARCHAR_TYPE:
	s += Chain(" VARCHAR(");
	s += Chain(dtlen);
	s += Chain(")");
	break;
    case BOOL_TYPE:
	s += Chain(" BOOL");
	break;
    case DATETIME_TYPE:
	s += Chain(" DATETIME");
	break;
    case BIGINT_TYPE:
	s += Chain(" bigint(");
	s += Chain(dtlen);
	s += Chain(")");
	break;
    case FLOAT_TYPE:
	s += Chain(" float");
	break;
    case DOUBLE_TYPE:
	s += Chain(" double");
	break;
    case DECIMAL_TYPE:
	s += Chain(" decimal(");
	s += Chain(dtlen);
	s += Chain(")");
	break;
    case SMALLINT_TYPE:
	s += Chain(" smallint");
	break;
    case TINYINT_TYPE:
	s += Chain(" tinyint");
	break;	  
    case NULL_TYPE:
	s += Chain(" null");
	break;
    case BLOB_TYPE:
    case CLOB_TYPE:
    case PAGEID_TYPE:
	throw Exception(EXLOC, Chain("Type not supported as return type"));
    }

    return s;
    
}

Chain OracleFormater::formatProcBlock(const ListT<CegoProcVar>& varList, const ListT<CegoProcStmt*>& stmtList,  const ListT<CegoProcException*>& exceptionList)
{
    Chain blockStmt;
    CegoProcVar* pVar = varList.First();
    while ( pVar )
    {
	blockStmt += Chain("DECLARE ") + pVar->dbFormat(this) + Chain(";\n");
	pVar = varList.Next();
    }
    
    return blockStmt;
}


Chain OracleFormater::formatProcThrowStmt(CegoExpr* pExpr)
{
    return Chain("TODO");
}

Chain OracleFormater::formatProcAssignStmt(CegoProcAssignStmt::AssignmentMode mode, const Chain paramName, CegoExpr* pExpr)
{
    return Chain("TODO");
}

Chain OracleFormater::formatProcBlockStmt(CegoProcBlock* pBlock)
{
    return Chain("TODO");
}

Chain OracleFormater::formatProcCursorCreateStmt(const Chain& name, CegoSelect* pSelect)
{
    return Chain("TODO");
}

Chain OracleFormater::formatProcCursorCloseStmt(const Chain& name)
{
    return Chain("TODO");
}

Chain OracleFormater::formatProcIfStmt(const ListT<CegoProcCond*>& condList, const ListT<CegoProcBlock*>& ifBlockList)
{
    return Chain("TODO");
}

Chain OracleFormater::formatProcNoopStmt()
{
    return Chain("TODO");
}

Chain OracleFormater::formatProcQueryStmt(CegoQuery *pQuery)
{
    return Chain("TODO");
}

Chain OracleFormater::formatProcReturnStmt(CegoExpr *pExpr)
{
    return Chain("TODO");
}


Chain OracleFormater::formatProcWhileStmt(CegoProcCond* pCond, CegoProcBlock* pBlock)
{
    return Chain("TODO");
}

Chain OracleFormater::formatProcCond(CegoProcCond::ConditionMode mode, CegoProcCond* pCond, CegoProcPred* pPred)
{
    return Chain("TODO");
}

Chain OracleFormater::formatProcPred(CegoProcPred::PredicateMode mode, CegoExpr* pExpr1, CegoExpr* pExpr2, CegoComparison comp, CegoProcCond* pCond, bool isNegateo)
{
    return Chain("TODO");
}

Chain OracleFormater::formatProcException(CegoException excep)
{
    return Chain("TODO");
}

Chain OracleFormater::formatExpr(CegoExpr::ExpType expType, CegoExpr* pExpr, CegoTerm* pTerm)
{
    return Chain("TODO");
}

Chain OracleFormater::formatTerm(CegoTerm::TermType termType, CegoTerm* pTerm, CegoFactor* pFactor)
{
    return Chain("TODO");
}

Chain OracleFormater::formatFactor(CegoFactor::FacType facType, const CegoFieldValue& fv, CegoAttrDesc* pAttrDesc, const Chain varName,
				  CegoExpr* pExpr, CegoProcFetch *pFetch, CegoAggregation* pAggr, CegoFunction* pFunction,
				  CegoSelect* pSelect, CegoCaseCond* pCaseCond )
{
    return Chain("TODO");
}

    
Chain OracleFormater::formatSelect(CegoSelect::SelectMode mode, bool isDistinct, const ListT<CegoExpr*>& exprList,
				   const ListT<CegoContentObject*>& coList, CegoPredDesc *pPred, ListT<CegoAttrDesc*> *pGroupList, CegoHavingCond* pHaving, ListT<CegoExpr*> *pOrderList, int rowLimit, CegoSelect* pUnionSelect)
{
    return Chain("TODO");
}

Chain OracleFormater::formatProcFetch(const Chain& curName, const ListT<Chain>& varList)
{
    return Chain("TODO");
}


Chain OracleFormater::formatHavingCond(CegoHavingCond::ConditionMode mode, CegoHavingCond* pCond, CegoHavingDesc* pDesc)
{

    Chain s;
    
    switch ( mode )
    {
    case CegoHavingCond::AND:
	s = pCond->dbFormat(this)  + Chain(" AND ") + pDesc->dbFormat(this);
	break;
    case CegoHavingCond::OR:
	s = pCond->dbFormat(this)  + Chain(" OR ") + pDesc->dbFormat(this);
	break;
    case CegoHavingCond::HAVINGDESC:
	s = pDesc->dbFormat(this);
	break;
    }
    return s;
}

Chain OracleFormater::formatHavingDesc(CegoExpr *pAggExpr, CegoExpr* pExpr, CegoComparison comp)
{
    Chain s;

    s += pAggExpr->dbFormat(this);
    
    switch ( comp )
    {
    case EQUAL:
	s += " = ";
	break;
    case NOT_EQUAL:
	s += " != ";
	break;
    case LESS_THAN:
	s += " < ";
	break;
    case MORE_THAN:
	s += " > ";
	break;
    case LESS_EQUAL_THAN:
	s += " <= ";
	break;
    case MORE_EQUAL_THAN:
	s += " >= ";
	break;
    }
    
    s += pExpr->dbFormat(this);
    
    return s;

}
    
Chain OracleFormater::formatPred(CegoPredDesc::CompMode mode, CegoExpr* pExpr1, CegoExpr *pExpr2, CegoExpr* pExpr3, ListT<CegoExpr*>& exprList, CegoComparison comp, const Chain& pattern, CegoCondDesc* pCond, CegoPredDesc* pPred, CegoSelect *pSelect)
{
    return Chain("TODO");
}

Chain OracleFormater::formatQuery(CegoQuery::QueryMode mode, const Chain& tableName, const Chain& tableAlias, const ListT<CegoField>& schema, const ListT<ListT<CegoExpr*> >& exprListArray,
				  const ListT<CegoExpr*>& updList, const ListT<CegoReturnVar*>& retVarList, const ListT<CegoAlterDesc>& alterList, CegoPredDesc *pPred, CegoSelect *pSelect,
				  const Chain& objName, CegoObject::ObjectType objType, const Chain& newObjName)
{
    return Chain("TODO");
}
    
Chain OracleFormater::formatAttr(const Chain& tableName, const Chain& attrName)
{
    return Chain("TODO");
}
    
Chain OracleFormater::formatFunction(CegoFunction::FunctionType type, const ListT<CegoExpr*>& exprList, const Chain& funcName, const Chain& counterId)
{
    return Chain("TODO");
}

Chain OracleFormater::formatAggregation(CegoAggregation::AggType type, CegoExpr *pExpr, bool isDistinct)
{
    return Chain("TODO");
}

Chain OracleFormater::formatCaseCond(const ListT<CegoPredDesc*>& predList, const ListT<CegoExpr*>& exprList, CegoExpr* elseExpr)
{
    return Chain("TODO");
}


Chain OracleFormater::formatReturnValue(const Chain& varName, CegoExpr* pExpr)
{
    return varName + Chain(" = ") + pExpr->dbFormat(this);
}

Chain OracleFormater::formatAlter(CegoAlterDesc::AlterType type, const CegoField& f, const Chain& attrName)
{
    return Chain("ALTER TODO");
}

Chain OracleFormater::formatFieldValue(CegoDataType type, void *pV, int len )
{
    Chain s;

    if ( type == NULL_TYPE || type == BLOB_TYPE )
    {
	s = Chain("null");
    }
    else
    {
	if ( pV == 0 )
	{
	    s = Chain("null");
	    return s;
	}
	switch ( type )
	{
	case INT_TYPE:
	{
	    int i;
	    memcpy(&i, pV, sizeof(int));
	    s = Chain( i );
	    break;
	}   
	case LONG_TYPE:
	{
	    long l;
	    memcpy(&l, pV, sizeof(long));
	    s = Chain( l );
	    break;
	}   
	case VARCHAR_TYPE:	    
	{   
	    Chain val((char*)pV);

	    // escape backslashes
	    Chain bval;
	    val.replaceAll(Chain("\\"), Chain("\\\\"), bval);

	    // escape newlines
	    Chain nval;
	    bval.replaceAll(Chain("\n"), Chain("\\n"), nval);

	    Chain qval;
	    nval.replaceAll(Chain("'"), Chain("\\'"), qval);

	    s = Chain("'") + qval + Chain("'");
	    break;
	}   
	case BOOL_TYPE:
	{   
	    char b = *(char*)(pV);
	    if ( b == 1 )
		s = Chain("true");
	    else
		s = Chain("false");
	    break;
	}    
	case DATETIME_TYPE:
	{
	    int i;
	    memcpy(&i, pV, sizeof(int));
	    if ( i == 0 )
	    {
		s = Chain("sysdate");
	    }
	    else
	    {
		Datetime dt ( i );
		s = Chain("'") + dt.asChain(ORACLEDATEFORMAT) + Chain("')");
	    }
	    break;
	}
	case BIGINT_TYPE:
	{
	    BigInteger bi( Chain((char*)pV) );
	    s =  bi.toChain();
	    break;
	}
	case DECIMAL_TYPE:
	{
	    BigDecimal d( Chain((char*)pV) );
	    s = d.toChain();
	    break;	    
	}
	case FLOAT_TYPE:
	{
	    float f;
	    memcpy(&f, pV, sizeof(float));
	    s = Chain( f );
	    break;
	}
	case DOUBLE_TYPE:
	{
	    double d;
	    memcpy(&d, pV, sizeof(double));
	    s = Chain( d );
	    break;
	}
	case SMALLINT_TYPE:
	{
	    short st;
	    memcpy(&st, pV, sizeof(short));
	    s = Chain( st );
	    break;
	}
	case TINYINT_TYPE:
	{
	    s = Chain( *(char*)(pV) );
	    break;
	}
	case BLOB_TYPE:
	case CLOB_TYPE:
	case PAGEID_TYPE:
	case NULL_TYPE:
	    // already handled
	    break;	    
	}
    }
    return s;

}

Chain OracleFormater::formatTrigger(const Chain& triggerName,
				    bool isBefore,
				    bool isOnInsert,
				    bool isOnUpdate,
				    bool isOnDelete,
				    const Chain& tableName,				
				    CegoProcBlock* pBlock)
{
    return Chain("TODO");    
}


